<?php

// TODO: fix unittests run by gitlab runner.

// Besprechen mit Carsten //
// ====================== //

// ?CR: QuickFormQuery.php L138: Why fallback $t3data[T3DATA_UID] = 0; and $t3data[T3DATA_BODYTEXT] = ''; ? When does this happen?
// ?CR: alle ?CR suchen und durchgehen.

namespace IMATHUZH\Qfq\Core\Report;

use IMATHUZH\Qfq\Core\Database\Database;
use IMATHUZH\Qfq\Core\Helper\HelperFile;
use IMATHUZH\Qfq\Core\Helper\Path;
use IMATHUZH\Qfq\Core\Helper\Sanitize;
use IMATHUZH\Qfq\Core\Store\Store;

class ReportAsFile
{
    /**
     * Finds the keyword file=<pathFileName> and returns pathFileName of report relative to CWD.
     *
     * @param string $bodyText
     * @return string|null pathFileName of report relative to CWD.
     * @throws \CodeException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public static function parseFileKeyword(string $bodyText) // : ?string
    {
        if (preg_match('/^\s*' . TOKEN_REPORT_FILE . '\s*=\s*([^\s]*)/m', $bodyText, $matches)) {
            return HelperFile::joinPathFilename(self::reportPath(), $matches[1]);
        } else {
            return null;
        }
    }

    /**
     * Read report file from given pathFileName. Throw exception on failure.
     *
     * @param string $reportPathFileNameFull
     * @return string|null
     * @throws \UserReportException
     */
    public static function read_report_file(string $reportPathFileNameFull): string
    {
        $fileContents = file_get_contents($reportPathFileNameFull);
        if ($fileContents === false) {
            throw new \UserReportException(json_encode([
                ERROR_MESSAGE_TO_USER => "File read error.",
                ERROR_MESSAGE_TO_DEVELOPER => "Report file not found or no permission to read file: '$reportPathFileNameFull'"]),
                ERROR_FORM_NOT_FOUND);
        }
        return $fileContents;
    }

    /**
     * Create new file named after the tt_content header or the UID if not defined.
     * - Invalid characters are stripped from filename.
     * - If file exists, throw exception.
     * - The path of the file is derived the typo3 page structure (use page alias if exists, else page name).
     * - The path is created if not exists.
     *
     * @param int $uid
     * @param Database $databaseT3
     * @return string pathFileName of report relative to CWD
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public static function create_file_from_ttContent(int $uid, Database $databaseT3): string
    {
        // Select tt_content
        $dbT3 = Store::getInstance()->getVar(SYSTEM_DB_NAME_T3, STORE_SYSTEM);
        $sql = "SELECT `bodytext`, `header`, `pid` FROM `$dbT3`.`tt_content` WHERE `uid` = ?";
        $ttContent = $databaseT3->sql($sql, ROW_EXPECT_1,
            [$uid], "Typo3 content element not found. Uid: $uid");

        // read page path of tt-content element (use page alias if exists)
        $reportPath = '';
        $pid = $ttContent['pid'];
        while (intval($pid) !== 0) {
            $sql = "SELECT `pid`, `alias`, `title` FROM `$dbT3`.`pages` WHERE `uid` = ?";
            $page = $databaseT3->sql($sql, ROW_EXPECT_1,
                [$pid], "Typo3 page not found. Uid: $pid");
            $supDir = $page['alias'] !== '' ? $page['alias'] : $page['title'];
            $reportPath = HelperFile::joinPathFilename(Sanitize::safeFilename($supDir), $reportPath);
            $pid = $page['pid'];
            if (strlen($reportPath) > 4096) {
                // Throw exception in case of infinite loop.
                throw new \UserReportException(json_encode([
                    ERROR_MESSAGE_TO_USER => "Path too long.",
                    ERROR_MESSAGE_TO_DEVELOPER => "Report path too long: '$reportPath'"]),
                    ERROR_FORM_NOT_FOUND);
            }
        }
        $reportPathFull = HelperFile::joinPathFilename(ReportAsFile::reportPath(), $reportPath);

        // create path in filesystem if not exists
        HelperFile::createPathRecursive($reportPathFull);

        // add filename
        $fileName = Sanitize::safeFilename($ttContent['header']);
        if (empty($fileName)) {
            $fileName = strval($uid);
        }
        $reportPathFileNameFull = HelperFile::joinPathFilename($reportPathFull, $fileName) . REPORT_FILE_EXTENSION;

        // if file exists, throw exception
        if (file_exists($reportPathFileNameFull)) {
            throw new \UserReportException(json_encode([
                ERROR_MESSAGE_TO_USER => "Writing file failed.",
                ERROR_MESSAGE_TO_DEVELOPER => "Can't export report to file. File already exists: '$reportPathFileNameFull'"]),
                ERROR_FORM_NOT_FOUND);
        }

        // create file with content bodytext
        HelperFile::file_put_contents($reportPathFileNameFull, $ttContent['bodytext']);

        // replace tt-content bodytext with file=<path>
        $newBodytext = TOKEN_REPORT_FILE . "=" . HelperFile::joinPathFilename($reportPath, $fileName) . REPORT_FILE_EXTENSION;
        $sql = "UPDATE `$dbT3`.`tt_content` SET `bodytext` = ?, `tstamp` = UNIX_TIMESTAMP(NOW()) WHERE `uid` = ?";
        $databaseT3->sql($sql, ROW_REGULAR, [$newBodytext, $uid]);

        // Clear cache
        // Need to truncate cf_cache_pages because it is used to restore page-specific cache
        $sql = "DELETE FROM `$dbT3`.`cf_cache_pages`";
        $databaseT3->sql($sql);

        return $reportPathFileNameFull;
    }

    /**
     * Overwrites report file given by the 'file=<path>' keyword of the tt_content element with the given uid.
     *
     * @param int $uid
     * @param string $newContent
     * @param Database $databaseT3
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public static function write_file_uid(int $uid, string $newContent, Database $databaseT3) // : void
    {
        // Get the report file path from tt_content
        $dbT3 = Store::getInstance()->getVar(SYSTEM_DB_NAME_T3, STORE_SYSTEM);
        $sql = "SELECT `bodytext` FROM `$dbT3`.`tt_content` WHERE `uid` = ?";
        $ttContent = $databaseT3->sql($sql, ROW_EXPECT_1,
            [$uid], "Typo3 content element not found. Uid: $uid");

        $reportPathFileNameFull = ReportAsFile::parseFileKeyword($ttContent['bodytext']);
        if ($reportPathFileNameFull === null) {
            throw new \UserReportException(json_encode([
                ERROR_MESSAGE_TO_USER => "No report file defined.",
                ERROR_MESSAGE_TO_DEVELOPER => "The keyword '" . TOKEN_REPORT_FILE . "' is not present in the typo3 content element with id $uid"]),
                ERROR_FORM_NOT_FOUND);
        }

        // Update report file
        HelperFile::file_put_contents($reportPathFileNameFull, $newContent);
    }

    /**
     * Return the path of the report directory relative to CWD.
     * Create path if it doesn't exist.
     *
     * @return string
     * @throws \UserFormException
     */
    private static function reportPath(): string
    {
        $cwdToReport = Path::cwdToProject(Path::PROJECT_DIR_TO_REPORT);
        HelperFile::createPathRecursive($cwdToReport);
        return $cwdToReport;
    }
}